from typing import Literal

LanguageCode = Literal[
    "abk",
    "ace",
    "ach",
    "ada",
    "ady",
    "afa",
    "afh",
    "afr",
    "ain",
    "aka",
    "akk",
    "alb",
    "sqi",
    "ale",
    "alg",
    "alt",
    "amh",
    "ang",
    "anp",
    "apa",
    "ara",
    "arc",
    "arg",
    "arm",
    "hye",
    "arn",
    "arp",
    "art",
    "arw",
    "asm",
    "ast",
    "ath",
    "aus",
    "ava",
    "ave",
    "awa",
    "aym",
    "aze",
    "bad",
    "bai",
    "bak",
    "bal",
    "bam",
    "ban",
    "baq",
    "eus",
    "bas",
    "bat",
    "bej",
    "bel",
    "bem",
    "ben",
    "ber",
    "bho",
    "bih",
    "bik",
    "bin",
    "bis",
    "bla",
    "bnt",
    "tib",
    "bod",
    "bos",
    "bra",
    "bre",
    "btk",
    "bua",
    "bug",
    "bul",
    "bur",
    "mya",
    "byn",
    "cad",
    "cai",
    "car",
    "cat",
    "cau",
    "ceb",
    "cel",
    "cze",
    "ces",
    "cha",
    "chb",
    "che",
    "chg",
    "chi",
    "zho",
    "chk",
    "chm",
    "chn",
    "cho",
    "chp",
    "chr",
    "chu",
    "chv",
    "chy",
    "cmc",
    "cnr",
    "cop",
    "cor",
    "cos",
    "cpe",
    "cpf",
    "cpp",
    "cre",
    "crh",
    "crp",
    "csb",
    "cus",
    "wel",
    "cym",
    "dak",
    "dan",
    "dar",
    "day",
    "del",
    "den",
    "ger",
    "deu",
    "dgr",
    "din",
    "div",
    "doi",
    "dra",
    "dsb",
    "dua",
    "dum",
    "dut",
    "nld",
    "dyu",
    "dzo",
    "efi",
    "egy",
    "eka",
    "gre",
    "ell",
    "elx",
    "eng",
    "enm",
    "epo",
    "est",
    "ewe",
    "ewo",
    "fan",
    "fao",
    "per",
    "fas",
    "fat",
    "fij",
    "fil",
    "fin",
    "fiu",
    "fon",
    "fre",
    "fra",
    "frm",
    "fro",
    "frr",
    "frs",
    "fry",
    "ful",
    "fur",
    "gaa",
    "gay",
    "gba",
    "gem",
    "geo",
    "kat",
    "gez",
    "gil",
    "gla",
    "gle",
    "glg",
    "glv",
    "gmh",
    "goh",
    "gon",
    "gor",
    "got",
    "grb",
    "grc",
    "grn",
    "gsw",
    "guj",
    "gwi",
    "hai",
    "hat",
    "hau",
    "haw",
    "heb",
    "her",
    "hil",
    "him",
    "hin",
    "hit",
    "hmn",
    "hmo",
    "hrv",
    "hsb",
    "hun",
    "hup",
    "iba",
    "ibo",
    "ice",
    "isl",
    "ido",
    "iii",
    "ijo",
    "iku",
    "ile",
    "ilo",
    "ina",
    "inc",
    "ind",
    "ine",
    "inh",
    "ipk",
    "ira",
    "iro",
    "ita",
    "jav",
    "jbo",
    "jpn",
    "jpr",
    "jrb",
    "kaa",
    "kab",
    "kac",
    "kal",
    "kam",
    "kan",
    "kar",
    "kas",
    "kau",
    "kaw",
    "kaz",
    "kbd",
    "kha",
    "khi",
    "khm",
    "kho",
    "kik",
    "kin",
    "kir",
    "kmb",
    "kok",
    "kom",
    "kon",
    "kor",
    "kos",
    "kpe",
    "krc",
    "krl",
    "kro",
    "kru",
    "kua",
    "kum",
    "kur",
    "kut",
    "lad",
    "lah",
    "lam",
    "lao",
    "lat",
    "lav",
    "lez",
    "lim",
    "lin",
    "lit",
    "lol",
    "loz",
    "ltz",
    "lua",
    "lub",
    "lug",
    "lui",
    "lun",
    "luo",
    "lus",
    "mac",
    "mkd",
    "mad",
    "mag",
    "mah",
    "mai",
    "mak",
    "mal",
    "man",
    "mao",
    "mri",
    "map",
    "mar",
    "mas",
    "may",
    "msa",
    "mdf",
    "mdr",
    "men",
    "mga",
    "mic",
    "min",
    "mis",
    "mkh",
    "mlg",
    "mlt",
    "mnc",
    "mni",
    "mno",
    "moh",
    "mon",
    "mos",
    "mul",
    "mun",
    "mus",
    "mwl",
    "mwr",
    "myn",
    "myv",
    "nah",
    "nai",
    "nap",
    "nau",
    "nav",
    "nbl",
    "nde",
    "ndo",
    "nds",
    "nep",
    "new",
    "nia",
    "nic",
    "niu",
    "nno",
    "nob",
    "nog",
    "non",
    "nor",
    "nqo",
    "nso",
    "nub",
    "nwc",
    "nya",
    "nym",
    "nyn",
    "nyo",
    "nzi",
    "oci",
    "oji",
    "ori",
    "orm",
    "osa",
    "oss",
    "ota",
    "oto",
    "paa",
    "pag",
    "pal",
    "pam",
    "pan",
    "pap",
    "pau",
    "peo",
    "phi",
    "phn",
    "pli",
    "pol",
    "pon",
    "por",
    "pra",
    "pro",
    "pus",
    "que",
    "raj",
    "rap",
    "rar",
    "roa",
    "roh",
    "rom",
    "rum",
    "ron",
    "run",
    "rup",
    "rus",
    "sad",
    "sag",
    "sah",
    "sai",
    "sal",
    "sam",
    "san",
    "sas",
    "sat",
    "scn",
    "sco",
    "sel",
    "sem",
    "sga",
    "sgn",
    "shn",
    "sid",
    "sin",
    "sio",
    "sit",
    "sla",
    "slo",
    "slk",
    "slv",
    "sma",
    "sme",
    "smi",
    "smj",
    "smn",
    "smo",
    "sms",
    "sna",
    "snd",
    "snk",
    "sog",
    "som",
    "son",
    "sot",
    "spa",
    "srd",
    "srn",
    "srp",
    "srr",
    "ssa",
    "ssw",
    "suk",
    "sun",
    "sus",
    "sux",
    "swa",
    "swe",
    "syc",
    "syr",
    "tah",
    "tai",
    "tam",
    "tat",
    "tel",
    "tem",
    "ter",
    "tet",
    "tgk",
    "tgl",
    "tha",
    "tig",
    "tir",
    "tiv",
    "tkl",
    "tlh",
    "tli",
    "tmh",
    "tog",
    "ton",
    "tpi",
    "tsi",
    "tsn",
    "tso",
    "tuk",
    "tum",
    "tup",
    "tur",
    "tut",
    "tvl",
    "twi",
    "tyv",
    "udm",
    "uga",
    "uig",
    "ukr",
    "umb",
    "und",
    "urd",
    "uzb",
    "vai",
    "ven",
    "vie",
    "vol",
    "vot",
    "wak",
    "wal",
    "war",
    "was",
    "wen",
    "wln",
    "wol",
    "xal",
    "xho",
    "yao",
    "yap",
    "yid",
    "yor",
    "ypk",
    "zap",
    "zbl",
    "zen",
    "zgh",
    "zha",
    "znd",
    "zul",
    "zun",
    "zxx",
    "zza",
]
"""ISO 639-2 alpha-3 language codes."""



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.partij_identificator import (
    CreatePartijIdentificatorData,
    ListPartijIdentificatorenParams,
    PartijIdentificator,
)


class PartijIdentificatorResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "partij-identificatoren"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.list_iter = self._make_list_iter(self.list)

    def list(
        self, *, params: ListPartijIdentificatorenParams | None = None
    ) -> PaginatedResponseBody[PartijIdentificator]:
        response = self._get(self.base_path, params=params)
        return cast(
            PaginatedResponseBody[PartijIdentificator], self.process_response(response)
        )

    def retrieve(self, /, uuid: str | uuid.UUID) -> PartijIdentificator:
        response = self._get(f"{self.base_path}/{str(uuid)}")
        return cast(PartijIdentificator, self.process_response(response))

    def create(
        self,
        *,
        data: CreatePartijIdentificatorData,
    ) -> PartijIdentificator:
        response = self._post(self.base_path, data=data)
        return cast(PartijIdentificator, self.process_response(response))



from typing import Any

from ape_pie import APIClient

from openklant_client._resources.actor import ActorResource
from openklant_client._resources.betrokkene import BetrokkeneResource
from openklant_client._resources.digitaal_adres import DigitaalAdresResource
from openklant_client._resources.interne_taak import InterneTaakResource
from openklant_client._resources.klant_contact import KlantContactResource
from openklant_client._resources.onderwerp_object import OnderwerpObjectResource
from openklant_client._resources.partij import PartijResource
from openklant_client._resources.partij_identificator import PartijIdentificatorResource


class OpenKlantClient(APIClient):
    partij: PartijResource
    partij_identificator: PartijIdentificatorResource
    digitaal_adres: DigitaalAdresResource
    klant_contact: KlantContactResource
    onderwerp_object: OnderwerpObjectResource
    actor: ActorResource
    interne_taak: InterneTaakResource
    betrokkene: BetrokkeneResource

    def __init__(
        self,
        base_url: str,
        *,
        token: str,
        request_kwargs: dict[str, Any] | None = None,
    ):
        if request_kwargs is None:
            request_kwargs = {}
        if "headers" not in request_kwargs:
            request_kwargs["headers"] = {}
        request_kwargs["headers"]["Authorization"] = f"Token {token}"

        super().__init__(base_url=base_url, request_kwargs=request_kwargs)

        self.partij = PartijResource(self)
        self.partij_identificator = PartijIdentificatorResource(self)
        self.digitaal_adres = DigitaalAdresResource(self)
        self.klant_contact = KlantContactResource(self)
        self.onderwerp_object = OnderwerpObjectResource(self)
        self.actor = ActorResource(self)
        self.interne_taak = InterneTaakResource(self)
        self.betrokkene = BetrokkeneResource(self)



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.klant_contact import (
    CreateKlantContactData,
    KlantContact,
    ListKlantContactParams,
    RetrieveKlantContactParams,
)


class KlantContactResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "klantcontacten"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.list_iter = self._make_list_iter(self.list)

    def create(
        self,
        *,
        data: CreateKlantContactData,
    ) -> KlantContact:
        response = self._post(self.base_path, data=data)
        return cast(KlantContact, self.process_response(response))

    def retrieve(
        self,
        /,
        uuid: str | uuid.UUID,
        *,
        params: RetrieveKlantContactParams | None = None,
    ) -> KlantContact:
        response = self._get(f"{self.base_path}/{str(uuid)}", params=params)
        return cast(KlantContact, self.process_response(response))

    def list(
        self, *, params: ListKlantContactParams | None = None
    ) -> PaginatedResponseBody[KlantContact]:
        response = self._get(f"{self.base_path}", params=params)
        return cast(
            PaginatedResponseBody[KlantContact],
            self.process_response(response),
        )



from typing import Literal, NotRequired, Required

from typing_extensions import TypedDict

# The API uses django-filter, which maps booleans to a BooleanField, of which
# the docs at
# https://django-filter.readthedocs.io/en/stable/ref/widgets.html#booleanwidget
# note:
# This widget converts its input into Python’s True/False values. It will
# convert all case variations of True and False into the internal Python values.
BooleanQueryParam = Literal["True", "False", "true", "false"] | bool


class CreateAdres(TypedDict):
    nummeraanduidingId: Required[str | None]
    adresregel1: NotRequired[str]
    adresregel2: NotRequired[str]
    adresregel3: NotRequired[str]
    land: NotRequired[str]
    huisnummer: int
    postcode: str
    huisnummertoevoeging: str


class Adres(TypedDict):
    nummeraanduidingId: str
    adresregel1: str
    adresregel2: str
    adresregel3: str
    huisnummer: int | None
    postcode: str
    huisnummertoevoeging: str
    land: str


class ForeignKeyRef(TypedDict):
    uuid: str


class FullForeigKeyRef(TypedDict):
    uuid: str
    url: str



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources import Partij
from openklant_client.types.resources.digitaal_adres import (
    DigitaalAdres,
)
from tests.factories.digitaal_adres import DigitaalAdresCreateDataFactory
from tests.factories.partij import CreatePartijPersoonDataFactory
from tests.validators import DigitaalAdresCreateDataValidator, DigitaalAdresValidator


@pytest.fixture()
def een_partij(client) -> Partij:
    data = CreatePartijPersoonDataFactory()
    return client.partij.create_persoon(data=data)


@pytest.fixture()
def een_digitaal_adres(client, een_partij) -> Partij:
    data = DigitaalAdresCreateDataFactory(
        verstrektDoorBetrokkene=None,
        verstrektDoorPartij__uuid=een_partij["uuid"],
    )
    return client.digitaal_adres.create(data=data)


@pytest.mark.vcr
def test_create_digitaal_adres(client, een_partij) -> None:
    data = DigitaalAdresCreateDataValidator.validate_python(
        {
            "adres": "foo@bar.com",
            "omschrijving": "professional",
            "soortDigitaalAdres": "email",
            "verstrektDoorBetrokkene": None,
            "verstrektDoorPartij": {"uuid": een_partij["uuid"]},
        }
    )
    resp = client.digitaal_adres.create(
        data=data,
    )

    DigitaalAdresValidator.validate_python(resp)


@pytest.mark.usefixtures("een_digitaal_adres")
@pytest.mark.vcr
def test_list_digitaal_adres(client) -> None:
    resp = client.digitaal_adres.list()
    TypeAdapter(PaginatedResponseBody[DigitaalAdres]).validate_python(resp)


@pytest.mark.vcr
def test_retrieve_digitaal_adres(client, een_digitaal_adres) -> None:
    resp = client.digitaal_adres.retrieve(een_digitaal_adres["uuid"])
    TypeAdapter(DigitaalAdres).validate_python(resp)



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.actor import (
    Actor,
    ActorListParams,
    CreateActorData,
)


class ActorResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "actoren"

    def create(
        self,
        *,
        data: CreateActorData,
    ) -> Actor:
        response = self._post(self.base_path, data=data)
        return cast(Actor, self.process_response(response))

    def retrieve(self, /, uuid: str | uuid.UUID) -> Actor:
        response = self._get(f"{self.base_path}/{str(uuid)}")
        return cast(Actor, self.process_response(response))

    def list(
        self, *, params: ActorListParams | None = None
    ) -> PaginatedResponseBody[Actor]:
        response = self._get(f"{self.base_path}", params=params)
        return cast(
            PaginatedResponseBody[Actor],
            self.process_response(response),
        )



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import Adres, ForeignKeyRef, FullForeigKeyRef

BetrokkeneRol = Literal["vertegenwoordiger", "klant"]


class CreateContactnaam(TypedDict):
    voorletters: NotRequired[str]
    voornaam: str
    voorvoegselAchternaam: NotRequired[str]
    achternaam: str


class BetrokkeneCreateData(TypedDict):
    wasPartij: ForeignKeyRef | None
    hadKlantcontact: ForeignKeyRef | None
    bezoekadres: NotRequired[Adres]
    correspondentieadres: NotRequired[Adres]
    contactnaam: NotRequired[CreateContactnaam | None]
    rol: BetrokkeneRol
    organisatienaam: str
    initiator: bool


class Betrokkene(TypedDict):
    uuid: str
    url: str
    wasPartij: FullForeigKeyRef
    hadKlantcontact: FullForeigKeyRef
    digitaleAdressen: list[FullForeigKeyRef]
    bezoekadres: Adres
    correspondentieadres: Adres
    contactnaam: CreateContactnaam
    volledigeNaam: str
    rol: BetrokkeneRol
    organisatienaam: str
    initiator: bool


class BetrokkeneRetrieveParams(TypedDict):
    expand: NotRequired[list[Literal["digitaleAdressen",]]]



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import BooleanQueryParam

SoortActor = Literal["medewerker", "geautomatiseerde_actor", "organisatorische_eenheid"]


class ActorIdentificator(TypedDict):
    objectId: str
    codeObjecttype: str
    codeRegister: str
    codeSoortObjectId: str


class CreateActorData(TypedDict):
    naam: str
    soortActor: SoortActor
    indicatieActief: NotRequired[bool]
    actoridentificator: NotRequired[ActorIdentificator]


class Actor(TypedDict):
    uuid: str
    naam: str
    soortActor: SoortActor
    indicatieActief: bool
    actoridentificator: ActorIdentificator


class ActorListParams(TypedDict, total=False):
    actoridentificatorCodeObjecttype: str
    actoridentificatorCodeRegister: str
    actoridentificatorCodeSoortObjectId: str
    actoridentificatorObjectId: str
    indicatieActief: BooleanQueryParam
    naam: str
    page: int
    soortActor: SoortActor



import pytest
from pydantic import TypeAdapter

from openklant_client.exceptions import BadRequest
from openklant_client.types import PaginatedResponseBody
from openklant_client.types.resources.partij import Partij
from tests.factories.partij import (
    CreatePartijContactPersoonDataFactory,
    CreatePartijOrganisatieDataFactory,
    CreatePartijPersoonDataFactory,
)
from tests.validators import (
    CreatePartijContactpersoonDataValidator,
    CreatePartijOrganisatieDataValidator,
    CreatePartijPersoonDataValidator,
    PartijValidator,
)


@pytest.fixture()
def een_persoon(client):
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "crp",
        "soortPartij": "persoon",
        "partijIdentificatie": {
            "contactnaam": {
                "voorletters": "Dr.",
                "voornaam": "Test Persoon",
                "voorvoegselAchternaam": "Mrs.",
                "achternaam": "Gamble",
            }
        },
    }
    return client.partij.create_persoon(data=data)


@pytest.fixture()
def een_organisatie(client):
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "tiv",
        "soortPartij": "organisatie",
        "partijIdentificatie": {"naam": "Test Organisatie"},
    }
    return client.partij.create_organisatie(data=data)


@pytest.mark.vcr
@pytest.mark.usefixtures("een_persoon", "een_organisatie")
def test_list_partijen(client):
    resp = client.partij.list()

    TypeAdapter(PaginatedResponseBody[Partij]).validate_python(resp)


@pytest.mark.vcr
@pytest.mark.parametrize(
    "expand",
    (
        "betrokkenen",
        "betrokkenen.hadKlantcontact",
        "categorieRelaties",
        "digitaleAdressen",
    ),
)
def test_retrieve_partij(client, een_persoon, expand):
    resp = client.partij.retrieve(
        een_persoon["uuid"],
        params={"expand": [expand]},
    )

    PartijValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_persoon(client) -> None:
    data = CreatePartijPersoonDataValidator.validate_python(
        {
            "soortPartij": "persoon",
            "digitaleAdressen": None,
            "rekeningnummers": None,
            "voorkeursRekeningnummer": None,
            "voorkeurstaal": "nld",
            "indicatieActief": True,
            "indicatieGeheimhouding": False,
            "voorkeursDigitaalAdres": None,
            "partijIdentificatie": {"contactnaam": None},
        }
    )
    resp = client.partij.create_persoon(
        data=data,
    )

    PartijValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_persoon_with_identificator(client) -> None:
    data = CreatePartijPersoonDataValidator.validate_python(
        {
            "soortPartij": "persoon",
            "digitaleAdressen": None,
            "rekeningnummers": None,
            "voorkeursRekeningnummer": None,
            "voorkeurstaal": "nld",
            "indicatieActief": True,
            "indicatieGeheimhouding": False,
            "voorkeursDigitaalAdres": None,
            "partijIdentificatie": {"contactnaam": None},
            "partijIdentificatoren": [
                {
                    "partijIdentificator": {
                        "codeObjecttype": "natuurlijk_persoon",
                        "codeSoortObjectId": "bsn",
                        "objectId": "631706549",
                        "codeRegister": "brp",
                    }
                }
            ],
        }
    )
    resp = client.partij.create_persoon(
        data=data,
    )

    PartijValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_contactpersoon(client, een_organisatie):
    data = CreatePartijContactpersoonDataValidator.validate_python(
        {
            "soortPartij": "contactpersoon",
            "digitaleAdressen": None,
            "rekeningnummers": None,
            "voorkeursRekeningnummer": None,
            "voorkeurstaal": "nld",
            "indicatieActief": True,
            "indicatieGeheimhouding": False,
            "voorkeursDigitaalAdres": None,
            "partijIdentificatie": {
                "contactnaam": None,
                "werkteVoorPartij": {"uuid": een_organisatie["uuid"]},
            },
        }
    )
    resp = client.partij.create_contactpersoon(
        data=data,
    )

    PartijValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_organisatie(client):
    data = CreatePartijOrganisatieDataValidator.validate_python(
        {
            "soortPartij": "organisatie",
            "digitaleAdressen": None,
            "rekeningnummers": None,
            "voorkeursRekeningnummer": None,
            "voorkeurstaal": "nld",
            "indicatieActief": True,
            "indicatieGeheimhouding": False,
            "voorkeursDigitaalAdres": None,
            "partijIdentificatie": {"naam": "AcmeCorp Ltd"},
        }
    )
    resp = client.partij.create_organisatie(
        data=data,
    )

    PartijValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_with_bad_request_exception(client):
    with pytest.raises(BadRequest) as exc_info:
        client.partij.create_organisatie(
            data={},
        )

    got = (
        exc_info.value.status,
        exc_info.value.code,
        exc_info.value.title,
        exc_info.value.invalidParams,
        str(exc_info.value),
    )
    want = (
        400,
        "invalid",
        "Invalid input.",
        [
            {
                "name": "digitaleAdressen",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
            {
                "name": "voorkeursDigitaalAdres",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
            {
                "name": "rekeningnummers",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
            {
                "name": "voorkeursRekeningnummer",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
            {
                "name": "soortPartij",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
            {
                "name": "indicatieActief",
                "code": "required",
                "reason": "Dit veld is vereist.",
            },
        ],
        (
            'status=400 code=400 title="Invalid input.":\n'
            "Invalid parameters:\n"
            "{'code': 'required',\n"
            " 'name': 'digitaleAdressen',\n"
            " 'reason': 'Dit veld is vereist.'}\n"
            "{'code': 'required',\n"
            " 'name': 'voorkeursDigitaalAdres',\n"
            " 'reason': 'Dit veld is vereist.'}\n"
            "{'code': 'required',\n"
            " 'name': 'rekeningnummers',\n"
            " 'reason': 'Dit veld is vereist.'}\n"
            "{'code': 'required',\n"
            " 'name': 'voorkeursRekeningnummer',\n"
            " 'reason': 'Dit veld is vereist.'}\n"
            "{'code': 'required', 'name': 'soortPartij', "
            "'reason': 'Dit veld is vereist.'}\n"
            "{'code': 'required',\n"
            " 'name': 'indicatieActief',\n"
            " 'reason': 'Dit veld is vereist.'}"
        ),
    )

    assert got == want


@pytest.mark.vcr
def test_partial_update(client, een_persoon):
    target_nummer = "18744"
    target_correspondentieadres = {
        "adresregel1": "foo",
        "adresregel2": "bar",
        "adresregel3": "baz",
        "huisnummer": 1874,
        "postcode": "1234 AB",
        "stad": "Amsterdam",
        "straatnaam": "Teleportweg",
        "land": "NL",
        "nummeraanduidingId": "",
        "huisnummertoevoeging": "",
    }
    assert een_persoon["nummer"] != target_nummer
    assert een_persoon["correspondentieadres"] != target_correspondentieadres

    resp = client.partij.partial_update(
        een_persoon["uuid"],
        data={
            # TODO: remove soortPartij which shouldn't be required, see:
            # https://github.com/maykinmedia/open-klant/issues/345
            "correspondentieadres": target_correspondentieadres,
            "soortPartij": een_persoon["soortPartij"],
            "nummer": target_nummer,
        },
    )

    PartijValidator.validate_python(resp)
    assert resp["nummer"] == target_nummer
    assert resp["correspondentieadres"] == target_correspondentieadres


def test_factory_partij_persoon_data():
    data = CreatePartijPersoonDataFactory.build()
    CreatePartijPersoonDataValidator.validate_python(data)


def test_factory_partij_organisatie_data():
    data = CreatePartijOrganisatieDataFactory.build()
    CreatePartijOrganisatieDataValidator.validate_python(data)


def test_factory_partij_contactpersoon_data():
    data = CreatePartijContactPersoonDataFactory.build()
    CreatePartijContactpersoonDataValidator.validate_python(data)



from typing import Required

from typing_extensions import TypedDict


class PaginatedResponseBody[T](TypedDict):
    count: Required[int]
    next: Required[str | None]
    previous: Required[str | None]
    results: Required[list[T]]



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.onderwerp_object import (
    CreateOnderwerpObjectData,
    OnderwerpObject,
)
from tests.factories.klant_contact import CreateKlantContactDataFactory
from tests.factories.onderwerp_object import (
    CreateOnderwerpObjectDataFactory,
)
from tests.validators import (
    CreateOnderwerpObjectDataValidator,
    OnderwerpObjectValidator,
)


@pytest.fixture()
def een_klant_contact(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def een_ander_klant_contact(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def klant_contact_factory(client):
    def factory(*args, **kwargs):
        data = CreateKlantContactDataFactory(*args, **kwargs)
        return client.klant_contact.create(data=data)

    return factory


@pytest.fixture()
def een_onderwerp_object(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def onderwerp_object_factory(client, klant_contact_factory):
    def factory():
        klant_contact = klant_contact_factory()
        nog_een_klant_contact = klant_contact_factory()
        data = CreateOnderwerpObjectDataFactory(
            klantcontact={"uuid": klant_contact["uuid"]},
            wasKlantcontact={"uuid": nog_een_klant_contact["uuid"]},
        )
        return client.onderwerp_object.create(data=data)

    return factory


@pytest.mark.vcr
def test_create_onderwerp_object(client, een_klant_contact, een_ander_klant_contact):
    data: CreateOnderwerpObjectData = {
        "klantcontact": {"uuid": een_klant_contact["uuid"]},
        "wasKlantcontact": {"uuid": een_ander_klant_contact["uuid"]},
    }
    CreateOnderwerpObjectDataValidator.validate_python(data)
    resp = client.onderwerp_object.create(data=data)

    OnderwerpObjectValidator.validate_python(resp)


@pytest.mark.vcr
def test_list_onderwerp_objecten(client, onderwerp_object_factory):
    onderwerp_objecten = [onderwerp_object_factory() for _ in range(5)]

    resp = client.onderwerp_object.list()
    TypeAdapter(PaginatedResponseBody[OnderwerpObject]).validate_python(resp)

    assert {result["uuid"] for result in resp["results"]} == {
        oo["uuid"] for oo in onderwerp_objecten
    }



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.betrokkene import (
    Betrokkene,
)
from openklant_client.types.resources.partij import Partij
from tests.factories.betrokkene import BetrokkeneCreateDataFactory
from tests.factories.klant_contact import CreateKlantContactDataFactory
from tests.factories.partij import CreatePartijPersoonDataFactory
from tests.validators import BetrokkeneCreateDataValidator, BetrokkeneValidator


@pytest.fixture()
def een_persoon(client) -> Partij:
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "crp",
        "soortPartij": "persoon",
        "partijIdentificatie": {
            "contactnaam": {
                "voorletters": "Dr.",
                "voornaam": "Test Persoon",
                "voorvoegselAchternaam": "Mrs.",
                "achternaam": "Gamble",
            }
        },
    }
    return client.partij.create_persoon(data=data)


@pytest.fixture()
def een_organisatie(client):
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "tiv",
        "soortPartij": "organisatie",
        "partijIdentificatie": {"naam": "Test Organisatie"},
    }
    return client.partij.create_organisatie(data=data)


@pytest.fixture()
def een_klant_contact(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def persoon_factory(client):
    def factory(*args, **kwargs):
        data = CreatePartijPersoonDataFactory(*args, **kwargs)
        return client.partij.create_persoon(data=data)

    return factory


@pytest.fixture()
def klantcontact_factory(client):
    def factory(*args, **kwargs):
        data = CreateKlantContactDataFactory(*args, **kwargs)
        return client.klant_contact.create(data=data)

    return factory


@pytest.fixture()
def betrokkene_factory(client, persoon_factory, klantcontact_factory):
    def factory(*args, **kwargs):
        klantcontact, partij = klantcontact_factory(), persoon_factory()

        data = BetrokkeneCreateDataFactory(
            *args,
            **kwargs,
            hadKlantcontact={"uuid": klantcontact["uuid"]},
            wasPartij={"uuid": partij["uuid"]},
        )
        return client.betrokkene.create(data=data)

    return factory


@pytest.fixture()
def een_betrokkene(betrokkene_factory):
    return betrokkene_factory()


@pytest.mark.vcr
def test_create_betrokkene(client, een_persoon, een_klant_contact) -> None:
    data = BetrokkeneCreateDataValidator.validate_python(
        {
            "rol": "klant",
            "initiator": True,
            "organisatienaam": "foobar",
            "wasPartij": {"uuid": een_persoon["uuid"]},
            "hadKlantcontact": {"uuid": een_klant_contact["uuid"]},
        }
    )
    resp = client.betrokkene.create(data=data)

    BetrokkeneValidator.validate_python(resp)


@pytest.mark.vcr
def test_list_betrokkenen(client, betrokkene_factory):
    betrokkenen = [betrokkene_factory() for _ in range(5)]
    resp = client.betrokkene.list()

    TypeAdapter(PaginatedResponseBody[Betrokkene]).validate_python(resp)
    assert {result["uuid"] for result in resp["results"]} == {
        betrokkene["uuid"] for betrokkene in betrokkenen
    }


@pytest.mark.vcr
def test_retrieve_betrokkene(client, een_betrokkene):
    resp = client.betrokkene.retrieve(een_betrokkene["uuid"])

    BetrokkeneValidator.validate_python(resp)
    assert resp["uuid"] == een_betrokkene["uuid"]


@pytest.mark.vcr
def test_list_betrokkenen_as_pagination_iter(client, betrokkene_factory):
    # We can't specify the pagesize, so we have to use the default 100 to
    # create more than 1 page of data
    betrokkenen = [betrokkene_factory() for _ in range(101)]
    assert client.betrokkene.list()["next"] is not None

    resp = list(client.betrokkene.list_iter())

    TypeAdapter(list[Betrokkene]).validate_python(resp)
    assert sorted(resp, key=lambda b: b["uuid"]) == sorted(
        (b | {"_expand": {}} for b in betrokkenen), key=lambda b: b["uuid"]
    )



from typing import NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import ForeignKeyRef, FullForeigKeyRef


class OnderwerpObjectIdentificator(TypedDict):
    objectId: str
    codeObjecttype: str
    codeRegister: str
    codeSoortObjectId: str


class BetrokkeneBase(TypedDict):
    uuid: str
    url: str


class CreateOnderwerpObjectData(TypedDict):
    wasKlantcontact: ForeignKeyRef | None
    klantcontact: ForeignKeyRef | None
    onderwerpobjectidentificator: NotRequired[OnderwerpObjectIdentificator | None]


class OnderwerpObject(TypedDict):
    uuid: str
    url: str
    wasKlantcontact: FullForeigKeyRef | None
    klantcontact: FullForeigKeyRef | None
    onderwerpobjectidentificator: OnderwerpObjectIdentificator


class OnderwerpobjectIdentificatorListParams(TypedDict, total=False):
    onderwerpobjectidentificatorCodeObjecttype: str
    onderwerpobjectidentificatorCodeRegister: str
    onderwerpobjectidentificatorCodeSoortObjectId: str
    onderwerpobjectidentificatorObjectId: str



from typing import Literal

from typing_extensions import TypedDict


class InvalidParam(TypedDict):
    name: str
    code: str
    reason: str


class ErrorResponseBody(TypedDict):
    type: str
    code: str
    title: str
    status: int
    detail: str
    instance: str


class ValidationErrorResponseBody(ErrorResponseBody):
    status: Literal[400]
    invalidParams: list[InvalidParam]



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.actor import Actor
from tests.factories.actor import CreateActorDataFactory
from tests.validators import ActorValidator, CreateActorDataValidator


@pytest.fixture()
def actor_factory(client):
    def factory(*args, **kwargs):
        data = CreateActorDataFactory(*args, **kwargs)
        return client.actor.create(data=data)

    return factory


@pytest.fixture()
def een_actor(actor_factory):
    return actor_factory()


@pytest.mark.vcr
def test_create_actor(client) -> None:
    data = CreateActorDataValidator.validate_python(
        {
            "naam": "Miranda Peters",
            "soortActor": "medewerker",
            "indicatieActief": True,
            "actoridentificator": {
                "objectId": "6070462185571",
                "codeObjecttype": "exist",
                "codeRegister": "sport",
                "codeSoortObjectId": "style",
            },
        }
    )
    resp = client.actor.create(data=data)

    ActorValidator.validate_python(resp)


@pytest.mark.vcr
def test_list_actoren(client, actor_factory):
    actoren = [actor_factory() for _ in range(5)]
    resp = client.actor.list()

    TypeAdapter(PaginatedResponseBody[Actor]).validate_python(resp)
    assert {result["uuid"] for result in resp["results"]} == {
        actor["uuid"] for actor in actoren
    }


@pytest.mark.vcr
def test_retrieve_actor(client, een_actor):
    resp = client.actor.retrieve(een_actor["uuid"])

    ActorValidator.validate_python(resp)
    assert resp["uuid"] == een_actor["uuid"]


@pytest.mark.vcr
@pytest.mark.parametrize("indicatie_actief", (True, False))
@pytest.mark.parametrize(
    "soort_actor", ("medewerker", "geautomatiseerde_actor", "organisatorische_eenheid")
)
def test_list_actoren_with_boolean_query_params(
    client, actor_factory, indicatie_actief, soort_actor
):
    excluded_actoren = [
        actor_factory(indicatieActief=not indicatie_actief) for _ in range(2)
    ]
    included_actor = actor_factory(
        indicatieActief=indicatie_actief, soortActor=soort_actor
    )
    resp = client.actor.list(
        params={"indicatieActief": indicatie_actief, "soortActor": soort_actor}
    )

    TypeAdapter(PaginatedResponseBody[Actor]).validate_python(resp)
    uuids = {result["uuid"] for result in resp["results"]}
    assert included_actor["uuid"] in uuids
    assert not any(na["uuid"] in uuids for na in excluded_actoren)



from .common import CreateAdres
from .error import (
    ErrorResponseBody,
    InvalidParam,
    ValidationErrorResponseBody,
)
from .iso_639_2 import LanguageCode
from .pagination import PaginatedResponseBody

__all__ = [
    "CreateAdres",
    "LanguageCode",
    "PaginatedResponseBody",
    "InvalidParam",
    "ErrorResponseBody",
    "ValidationErrorResponseBody",
]



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import Adres, CreateAdres, ForeignKeyRef
from openklant_client.types.iso_639_2 import LanguageCode
from openklant_client.types.resources.digitaal_adres import DigitaalAdres
from openklant_client.types.resources.partij_identificator import (
    CreateEmbeddedPartijIdentificatorData,
)

#
# Input types
#

SoortPartij = Literal["persoon", "contactpersoon", "organisatie"]


class CreateContactnaam(TypedDict):
    voorletters: str
    voornaam: str
    voorvoegselAchternaam: str
    achternaam: str


# Note this is polymorphic, concrete types below
class CreatePartijDataBase(TypedDict):
    soortPartij: Literal["organisatie", "persoon", "contactpersoon"]
    indicatieActief: bool
    nummer: NotRequired[str]
    interneNotitie: NotRequired[str]
    digitaleAdressen: list[ForeignKeyRef] | None
    voorkeursDigitaalAdres: ForeignKeyRef | None
    rekeningnummers: list[ForeignKeyRef] | None
    voorkeursRekeningnummer: ForeignKeyRef | None
    voorkeurstaal: LanguageCode
    indicatieGeheimhouding: bool
    correspondentieadres: NotRequired[CreateAdres]
    bezoekadres: NotRequired[CreateAdres | None]
    partijIdentificatoren: NotRequired[list[CreateEmbeddedPartijIdentificatorData]]


class CreatePartijIdentificatiePersoon(TypedDict):
    contactnaam: CreateContactnaam | None


class CreatePartijIdentificatieContactpersoon(TypedDict):
    contactnaam: CreateContactnaam | None
    werkteVoorPartij: ForeignKeyRef


class CreatePartijIdentificatieOrganisatie(TypedDict):
    naam: str


class CreatePartijPersoonData(CreatePartijDataBase):
    soortPartij: Literal["persoon"]
    partijIdentificatie: CreatePartijIdentificatiePersoon


class CreatePartijContactpersoonData(CreatePartijDataBase):
    soortPartij: Literal["contactpersoon"]
    partijIdentificatie: CreatePartijIdentificatieContactpersoon


class CreatePartijOrganisatieData(CreatePartijDataBase):
    soortPartij: Literal["organisatie"]
    partijIdentificatie: CreatePartijIdentificatieOrganisatie


class PartialUpdatePartijData(CreatePartijDataBase, total=False):
    # TODO: This is required for PATCH, which is a bug. Should be remove when
    # https://github.com/maykinmedia/open-klant/issues/345 is addressed.
    soortPartij: SoortPartij


class PartijListParams(TypedDict, total=False):
    page: int
    vertegenwoordigdePartij__url: str
    partijIdentificator__codeObjecttype: str
    partijIdentificator__codeRegister: str
    partijIdentificator__codeSoortObjectId: str
    partijIdentificator__objectId: str
    soortPartij: Literal["organisatie", "persoon", "contactpersoon"]
    expand: NotRequired[
        list[
            Literal[
                "betrokkenen",
                "betrokkenen.hadKlantcontact",
                "categorieRelaties",
                "digitaleAdressen",
            ]
        ]
    ]


class PartijRetrieveParams(TypedDict):
    expand: NotRequired[
        list[
            Literal[
                "betrokkenen",
                "betrokkenen.hadKlantcontact",
                "categorieRelaties",
                "digitaleAdressen",
            ]
        ]
    ]


#
# Output types
#


class Contactnaam(TypedDict):
    voorletters: str
    voornaam: str
    voorvoegselAchternaam: str
    achternaam: str


class PartijIdentificatiePersoon(TypedDict):
    contactnaam: Contactnaam | None


class PartijIdentificatieOrganisatie(TypedDict):
    naam: str


class PartijIdentificatieContactpersoon(TypedDict):
    contactnaam: Contactnaam | None


class PartijExpand(TypedDict, total=False):
    digitaleAdressen: list[DigitaalAdres]
    # TODO: betrokkenen, categorie_relaties


class Partij(TypedDict):
    uuid: str
    nummer: str
    url: str
    interneNotitie: str
    digitaleAdressen: list[ForeignKeyRef]
    voorkeursDigitaalAdres: ForeignKeyRef | None
    rekeningnummers: list[ForeignKeyRef]
    voorkeursRekeningnummer: ForeignKeyRef | None
    voorkeurstaal: LanguageCode
    indicatieActief: bool
    indicatieGeheimhouding: bool
    correspondentieadres: Adres | None
    bezoekadres: Adres | None
    soortPartij: Literal["organisatie", "persoon", "contactpersoon"]
    partijIdentificatie: (
        PartijIdentificatieContactpersoon
        | PartijIdentificatiePersoon
        | PartijIdentificatieOrganisatie
    )
    _expand: NotRequired[PartijExpand]


#
# Validators
#



from .partij import (
    CreateContactnaam,
    CreatePartijContactpersoonData,
    CreatePartijDataBase,
    CreatePartijIdentificatieContactpersoon,
    CreatePartijIdentificatieOrganisatie,
    CreatePartijIdentificatiePersoon,
    CreatePartijOrganisatieData,
    CreatePartijPersoonData,
    Partij,
    PartijListParams,
    PartijRetrieveParams,
)

__all__ = [
    "CreateContactnaam",
    "CreatePartijDataBase",
    "CreatePartijIdentificatiePersoon",
    "CreatePartijIdentificatieContactpersoon",
    "CreatePartijIdentificatieOrganisatie",
    "CreatePartijPersoonData",
    "CreatePartijContactpersoonData",
    "CreatePartijOrganisatieData",
    "Partij",
    "PartijListParams",
    "PartijRetrieveParams",
]



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.interne_taak import (
    InterneTaak,
)
from tests.factories.actor import CreateActorDataFactory
from tests.factories.interne_taak import CreateInterneTaakFactory
from tests.factories.klant_contact import CreateKlantContactDataFactory
from tests.validators import CreateInterneTaakDataValidator, InterneTaakValidator


@pytest.fixture()
def een_klant_contact(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def een_actor(client):
    data = CreateActorDataFactory()
    return client.actor.create(data=data)


@pytest.fixture()
def interne_taak_factory(client, een_klant_contact, een_actor):
    def factory(*args, **kwargs):
        data = CreateInterneTaakFactory(
            *args,
            **kwargs,
            aanleidinggevendKlantcontact={"uuid": een_klant_contact["uuid"]},
            toegewezenAanActor={"uuid": een_actor["uuid"]},
        )
        return client.interne_taak.create(data=data)

    return factory


@pytest.fixture()
def een_interne_taak(interne_taak_factory):
    return interne_taak_factory()


@pytest.mark.vcr
def test_create_interne_taak(client, een_klant_contact, een_actor) -> None:
    data = CreateInterneTaakDataValidator.validate_python(
        {
            "gevraagdeHandeling": "Foobar",
            "aanleidinggevendKlantcontact": {"uuid": een_klant_contact["uuid"]},
            "toegewezenAanActor": {"uuid": een_actor["uuid"]},
            "toelichting": "Dit is een vraag",
            "status": "te_verwerken",
        }
    )
    resp = client.interne_taak.create(data=data)
    InterneTaakValidator.validate_python(resp)


@pytest.mark.vcr
def test_list_interne_taken(client, interne_taak_factory) -> None:
    interne_taken = [interne_taak_factory() for _ in range(5)]
    resp = client.interne_taak.list()

    TypeAdapter(PaginatedResponseBody[InterneTaak]).validate_python(resp)
    assert {result["uuid"] for result in resp["results"]} == {
        it["uuid"] for it in interne_taken
    }


@pytest.mark.vcr
def test_retrieve_interne_taak(client, een_interne_taak) -> None:
    resp = client.interne_taak.retrieve(een_interne_taak["uuid"])

    InterneTaakValidator.validate_python(resp)
    assert resp["uuid"] == een_interne_taak["uuid"]



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.digitaal_adres import (
    DigitaalAdres,
    DigitaalAdresCreateData,
    ListDigitaalAdresParams,
)


class DigitaalAdresResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "digitaleadressen"

    def list(
        self, *, params: ListDigitaalAdresParams | None = None
    ) -> PaginatedResponseBody[DigitaalAdres]:
        response = self._get(self.base_path, params=params)
        return cast(
            PaginatedResponseBody[DigitaalAdres], self.process_response(response)
        )

    def retrieve(self, /, uuid: str | uuid.UUID) -> DigitaalAdres:
        response = self._get(f"{self.base_path}/{str(uuid)}")
        return cast(DigitaalAdres, self.process_response(response))

    def create(
        self,
        *,
        data: DigitaalAdresCreateData,
    ) -> DigitaalAdres:
        response = self._post(self.base_path, data=data)
        return cast(DigitaalAdres, self.process_response(response))

    def delete(self, /, uuid: str):
        return self._delete(f"{self.base_path}/{str(uuid)}")



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import ForeignKeyRef, FullForeigKeyRef

SoortDigitaalAdres = Literal["email", "telefoonnummer", "overig"]


#
# Input
#


class DigitaalAdresCreateData(TypedDict):
    verstrektDoorBetrokkene: ForeignKeyRef | None
    verstrektDoorPartij: ForeignKeyRef | None
    adres: str
    omschrijving: str
    soortDigitaalAdres: SoortDigitaalAdres
    isStandaardAdres: NotRequired[bool]


class ListDigitaalAdresParams(TypedDict):
    page: NotRequired[int]
    verstrektDoorPartij__uuid: NotRequired[str]
    verstrektDoorBetrokkene__uuid: NotRequired[str]


#
# Output
#


class DigitaalAdres(TypedDict):
    uuid: str
    url: str
    verstrektDoorBetrokkene: FullForeigKeyRef | None
    verstrektDoorPartij: FullForeigKeyRef | None
    adres: str
    omschrijving: str
    soortDigitaalAdres: SoortDigitaalAdres
    isStandaardAdres: bool



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import ForeignKeyRef, FullForeigKeyRef


class CreateInterneTaakData(TypedDict):
    nummer: NotRequired[str]
    gevraagdeHandeling: str
    aanleidinggevendKlantcontact: ForeignKeyRef
    toegewezenAanActor: ForeignKeyRef
    toelichting: NotRequired[str]
    status: Literal["te_verwerken", "verwerkt"]


class InterneTaak(TypedDict):
    uuid: str
    url: str
    nummer: str
    gevraagdeHandeling: str
    aanleidinggevendKlantcontact: FullForeigKeyRef
    toegewezenAanActor: FullForeigKeyRef
    toelichting: str
    status: Literal["te_verwerken", "verwerkt"]
    toegewezenOp: str
    afgehandeldOp: str | None



from __future__ import annotations

import json
import logging
from collections.abc import Callable, Generator, Mapping, MutableMapping
from typing import (
    TYPE_CHECKING,
    Any,
    ParamSpec,
    TypeGuard,
    TypeVar,
    cast,
)

import requests
from ape_pie import APIClient

from openklant_client.exceptions import (
    BadRequest,
    Forbidden,
    InvalidJSONResponse,
    NonJSONResponse,
    NotFound,
    ResponseError,
    StructuredErrorResponse,
    Unauthorized,
)
from openklant_client.types.pagination import PaginatedResponseBody

if TYPE_CHECKING:
    JSONPrimitive = str | int | None | float
    JSONValue = JSONPrimitive | "JSONObject" | list["JSONValue"]
    JSONObject = dict[str, JSONValue]

logger = logging.getLogger(__name__)

ResourceResponse = MutableMapping[str, Any]

P = ParamSpec("P")
T = TypeVar("T")


def _is_error_response(data: Any) -> bool:
    """Check if response data matches error response structure."""
    if not isinstance(data, dict):
        return False
    required_keys = {"type", "code", "title", "status", "detail", "instance"}
    return required_keys.issubset(data.keys())


def _is_validation_error_response(data: Any) -> bool:
    """Check if response data matches validation error response structure."""
    if not _is_error_response(data):
        return False
    return data.get("status") == 400 and "invalidParams" in data


class ResourceMixin:
    http_client: APIClient

    def __init__(self, http_client: APIClient):
        self.http_client = http_client

    @staticmethod
    def process_response(response: requests.Response) -> TypeGuard[JSONValue]:
        response_data = None
        try:
            content_type = response.headers.get("Content-Type", "")
            # Note: there are currently no non-JSON responses defined in the
            # spec, the obvious example would be e.g. something like a blob
            # download. Until such endpoints are encountered, we treat non-JSON
            # as an error.
            if not content_type.lower().startswith("application/json"):
                raise NonJSONResponse(response)

            response_data = response.json()
        except (requests.exceptions.JSONDecodeError, json.JSONDecodeError) as exc:
            raise InvalidJSONResponse(response) from exc

        match response.status_code:
            case code if code >= 200 and code < 300 and response_data:
                return response_data
            case code if code >= 400 and code < 500 and response_data:
                exc_class = StructuredErrorResponse
                is_valid_error = _is_error_response(response_data)

                match code:
                    case 400:
                        if _is_validation_error_response(response_data):
                            raise BadRequest(response, response_data)
                        exc_class = BadRequest
                    case 401:
                        exc_class = Unauthorized
                    case 403:
                        exc_class = Forbidden
                    case 404:
                        exc_class = NotFound
                    case _:
                        pass

                if is_valid_error:
                    raise exc_class(response, response_data)
                # JSON body, but not in expected schema. Fall through to
                # generic ErrorResponse
            case _:
                pass

        raise ResponseError(response, msg="Error response")

    @staticmethod
    def _process_params(params: Mapping | None) -> None | Mapping:
        # The default approach to serializing lists in the requests library is
        # not supported by OpenKlant 2. See:
        # https://github.com/maykinmedia/open-klant/issues/250
        if params is None:
            return params

        transposed_params = dict(params)
        for key, val in params.items():
            if isinstance(val, list):
                transposed_params[key] = ",".join(str(element) for element in val)

        return transposed_params

    def _paginator(
        self,
        paginated_data: PaginatedResponseBody[T],
        max_requests: int | None = None,
    ) -> Generator[T, Any, None]:
        def row_iterator(
            _data: PaginatedResponseBody[T], num_requests=0
        ) -> Generator[T, Any, None]:
            for result in _data["results"]:
                yield cast(T, result)

            if next_url := _data.get("next"):
                if max_requests and num_requests >= max_requests:
                    logger.info(
                        "Number of requests while retrieving paginated "
                        "results reached maximum (%s), returning results. "
                        "Next URL: %s",
                        max_requests,
                        next_url,
                    )
                    return

                response = self.http_client.get(next_url)
                num_requests += 1
                response.raise_for_status()
                data = response.json()

                yield from row_iterator(data, num_requests)

        return row_iterator(paginated_data)

    def _get(
        self,
        path: str,
        headers: Mapping | None = None,
        params: Mapping | None = None,
    ) -> requests.Response:
        return self.http_client.request(
            "get", path, headers=headers, params=self._process_params(params)
        )

    def _post(
        self,
        path: str,
        headers: Mapping | None = None,
        params: Mapping | None = None,
        data: Any = None,
    ) -> requests.Response:
        return self.http_client.request(
            "post",
            path,
            headers=headers,
            json=data,
            params=self._process_params(params),
        )

    def _put(
        self,
        path: str,
        headers: Mapping | None = None,
        params: Mapping | None = None,
        data: Any = None,
    ) -> requests.Response:
        return self.http_client.request(
            "put", path, headers=headers, json=data, params=self._process_params(params)
        )

    def _delete(
        self,
        path: str,
        headers: Mapping | None = None,
        params: Mapping | None = None,
    ) -> requests.Response:
        return self.http_client.request(
            "delete",
            path,
            headers=headers,
            params=self._process_params(params),
        )

    def _patch(
        self,
        path: str,
        headers: Mapping | None = None,
        params: Mapping | None = None,
        data: Any = None,
    ) -> requests.Response:
        return self.http_client.request(
            "patch",
            path,
            headers=headers,
            params=self._process_params(params),
            json=data,
        )

    def _options(
        self,
        path: str,
        headers: Mapping | None = None,
        params: MutableMapping | None = None,
    ) -> requests.Response:
        return self.http_client.request(
            "delete",
            path,
            headers=headers,
            params=self._process_params(params),
        )

    def _make_list_iter(
        self, f: Callable[P, PaginatedResponseBody[T]]
    ) -> Callable[P, Generator[T, Any, Any]]:
        """Create a fully paginated iterator for the resource list() method."""

        def inner(*args: P.args, **kwargs: P.kwargs) -> Generator[T, Any, None]:
            return self._paginator(f(*args, **kwargs))

        return inner



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.klant_contact import (
    KlantContact,
)
from tests.factories.klant_contact import CreateKlantContactDataFactory
from tests.validators import CreateKlantContactDataValidator, KlantContactValidator


@pytest.fixture()
def een_klant_contact(client):
    data = CreateKlantContactDataFactory()
    return client.klant_contact.create(data=data)


@pytest.fixture()
def klant_contact_factory(client):
    def factory(*args, **kwargs):
        data = CreateKlantContactDataFactory(*args, **kwargs)
        return client.klant_contact.create(data=data)

    return factory


@pytest.mark.vcr
def test_create_klant_contact(client) -> None:
    data = CreateKlantContactDataValidator.validate_python(
        {
            "kanaal": "contactformulier",
            "onderwerp": "vraag",
            "inhoud": "Dit is een vraag",
            "taal": "nld",
            "vertrouwelijk": True,
        }
    )
    resp = client.klant_contact.create(data=data)

    KlantContactValidator.validate_python(resp)


@pytest.mark.vcr
def test_retrieve_klant_contact(client, een_klant_contact) -> None:
    resp = client.klant_contact.retrieve(een_klant_contact["uuid"])
    KlantContactValidator.validate_python(resp)


@pytest.mark.vcr
def test_list_klant_contacten(client, een_klant_contact) -> None:
    resp = client.klant_contact.list()
    TypeAdapter(PaginatedResponseBody[KlantContact]).validate_python(resp)
    assert [kc["uuid"] for kc in resp["results"]] == [een_klant_contact["uuid"]]


@pytest.mark.vcr
@pytest.mark.parametrize(
    "true",
    (True, "True", "true", "TRUE"),
)
@pytest.mark.parametrize(
    "false",
    (False, "False", "false", "FALSE", None),
)
def test_list_klant_contacten_with_boolean_filter_params(
    client, klant_contact_factory, true, false
) -> None:
    klant_contact = klant_contact_factory(indicatieContactGelukt=False)

    resp = client.klant_contact.list(
        params={
            "page": 1,
            "indicatieContactGelukt": false,
        }
    )
    TypeAdapter(PaginatedResponseBody[KlantContact]).validate_python(resp)
    assert [kc["uuid"] for kc in resp["results"]] == [klant_contact["uuid"]]

    resp = client.klant_contact.list(
        params={
            "page": 1,
            "indicatieContactGelukt": true,
        }
    )
    TypeAdapter(PaginatedResponseBody[KlantContact]).validate_python(resp)
    assert resp == {"next": None, "previous": None, "count": 0, "results": []}



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.onderwerp_object import (
    CreateOnderwerpObjectData,
    OnderwerpObject,
    OnderwerpobjectIdentificatorListParams,
)


class OnderwerpObjectResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "onderwerpobjecten"

    def create(
        self,
        *,
        data: CreateOnderwerpObjectData,
    ) -> OnderwerpObject:
        response = self._post(self.base_path, data=data)
        return cast(OnderwerpObject, self.process_response(response))

    def retrieve(self, /, uuid: str | uuid.UUID) -> OnderwerpObject:
        response = self._get(f"{self.base_path}/{str(uuid)}")
        return cast(OnderwerpObject, self.process_response(response))

    def list(
        self, *, params: OnderwerpobjectIdentificatorListParams | None = None
    ) -> PaginatedResponseBody[OnderwerpObject]:
        response = self._get(f"{self.base_path}", params=params)
        return cast(
            PaginatedResponseBody[OnderwerpObject],
            self.process_response(response),
        )



import pytest
from pydantic import TypeAdapter

from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources import Partij
from openklant_client.types.resources.partij_identificator import (
    PartijIdentificator,
)
from tests.validators import (
    CreatePartijIdentificatorDataValidator,
    PartijIdentificatorValidator,
)


@pytest.fixture()
def een_persoon(client) -> Partij:
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "crp",
        "soortPartij": "persoon",
        "partijIdentificatie": {
            "contactnaam": {
                "voorletters": "Dr.",
                "voornaam": "Test Persoon",
                "voorvoegselAchternaam": "Mrs.",
                "achternaam": "Gamble",
            }
        },
    }
    return client.partij.create_persoon(data=data)


@pytest.fixture()
def een_organisatie(client):
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "tiv",
        "soortPartij": "organisatie",
        "partijIdentificatie": {"naam": "Test Organisatie"},
    }
    return client.partij.create_organisatie(data=data)


@pytest.fixture()
def een_andere_organisatie(client):
    data = {
        "digitaleAdressen": None,
        "voorkeursDigitaalAdres": None,
        "rekeningnummers": None,
        "voorkeursRekeningnummer": None,
        "indicatieGeheimhouding": False,
        "indicatieActief": True,
        "voorkeurstaal": "tiv",
        "soortPartij": "organisatie",
        "partijIdentificatie": {"naam": "Andere Test Organisatie"},
    }
    return client.partij.create_organisatie(data=data)


@pytest.fixture()
def een_partij_identificator(client, een_persoon) -> Partij:
    data = {
        "identificeerdePartij": {"uuid": een_persoon["uuid"]},
        "partijIdentificator": {
            "codeObjecttype": "natuurlijk_persoon",
            "codeSoortObjectId": "bsn",
            "objectId": "631706549",
            "codeRegister": "brp",
        },
        "anderePartijIdentificator": "optional_identifier_123",
    }
    return client.partij_identificator.create(data=data)


@pytest.fixture()
def een_kvk_partij_identificator(client, een_organisatie) -> Partij:
    data = {
        "identificeerdePartij": {"uuid": een_organisatie["uuid"]},
        "partijIdentificator": {
            "codeObjecttype": "niet_natuurlijk_persoon",
            "codeSoortObjectId": "kvk_nummer",
            "objectId": "68750110",
            "codeRegister": "hr",
        },
        "anderePartijIdentificator": "optional_identifier_123",
    }
    return client.partij_identificator.create(data=data)


@pytest.mark.vcr
def test_create_partij_identificator(client, een_persoon) -> None:
    data = CreatePartijIdentificatorDataValidator.validate_python(
        {
            "identificeerdePartij": {"uuid": een_persoon["uuid"]},
            "partijIdentificator": {
                "codeObjecttype": "natuurlijk_persoon",
                "codeSoortObjectId": "bsn",
                "objectId": "631706549",
                "codeRegister": "brp",
            },
            "anderePartijIdentificator": "optional_identifier_123",
        }
    )
    resp = client.partij_identificator.create(
        data=data,
    )

    PartijIdentificatorValidator.validate_python(resp)


@pytest.mark.vcr
def test_create_vestiging_identificator(
    client, een_kvk_partij_identificator, een_andere_organisatie
) -> None:
    data = CreatePartijIdentificatorDataValidator.validate_python(
        {
            "identificeerdePartij": {"uuid": een_andere_organisatie["uuid"]},
            "partijIdentificator": {
                "codeObjecttype": "natuurlijk_persoon",
                "codeSoortObjectId": "bsn",
                "objectId": "631706549",
                "codeRegister": "brp",
            },
            "anderePartijIdentificator": "optional_identifier_123",
            "subIdentificatorVan": {
                "uuid": een_kvk_partij_identificator["uuid"],
            },
        }
    )
    resp = client.partij_identificator.create(
        data=data,
    )

    PartijIdentificatorValidator.validate_python(resp)


@pytest.mark.usefixtures("een_partij_identificator")
@pytest.mark.vcr
def test_list_partij_identificator(client) -> None:
    resp = client.partij_identificator.list()
    TypeAdapter(PaginatedResponseBody[PartijIdentificator]).validate_python(resp)


@pytest.mark.vcr
def test_retrieve_partij_identificator(client, een_partij_identificator) -> None:
    resp = client.partij_identificator.retrieve(een_partij_identificator["uuid"])
    TypeAdapter(PartijIdentificator).validate_python(resp)



import pprint
from typing import Literal, cast

from requests import Response

from openklant_client.types.error import (
    ErrorResponseBody,
    InvalidParam,
    ValidationErrorResponseBody,
)


class OpenKlant2Exception(Exception):
    """Base exception for all client-related errors."""

    pass


class ResponseError(OpenKlant2Exception):
    """A response was received, but it was erroneous in some aspect."""

    response: Response

    def __init__(self, response: Response, msg: str):
        self.response = response
        super().__init__(msg)


class NonJSONResponse(ResponseError):
    """Response was unexpectedly not in JSON."""

    def __init__(self, response: Response):
        content_type = response.headers.get("Content-Type")
        super().__init__(
            response,
            f"The content type of this response was {content_type}, but "
            "application/json was expected.",
        )


class InvalidJSONResponse(ResponseError):
    """Response was marked as JSON but contained malformed content."""

    def __init__(self, response: Response):
        super().__init__(
            response,
            "The response content-type was marked as JSON, but cannot be decoded",
        )


class StructuredErrorResponse(ResponseError):
    """An error response with a well-known JSON object describing the error."""

    type: str
    code: str
    title: str
    status: int
    detail: str
    instance: str

    def __init__(self, response: Response, body: ErrorResponseBody):
        self.type = body["type"]
        self.code = body["code"]
        self.title = body["title"]
        self.status = body["status"]
        self.detail = body["detail"]
        self.instance = body["instance"]

        super().__init__(
            response,
            f'status={self.status} code={self.code} title="{self.title}"',
        )


class ClientError(StructuredErrorResponse):
    pass


class BadRequest(ClientError):
    code: Literal[400]
    invalidParams: list[InvalidParam]

    def __init__(self, response: Response, body: ValidationErrorResponseBody):
        self.invalidParams = body["invalidParams"]

        invalid_params_formatted = ""
        if self.invalidParams:
            invalid_params_formatted = "\nInvalid parameters:\n"
            invalid_params_formatted += "\n".join(
                pprint.pformat(param) for param in self.invalidParams
            )

        StructuredErrorResponse.__init__(
            self,
            response,
            cast(ErrorResponseBody, body),
        )
        msg = (
            f"status={body['status']} code={body['status']} "
            f'title="{body["title"]}":{invalid_params_formatted}'
        )
        OpenKlant2Exception.__init__(self, msg)


class NotFound(ClientError):
    code: Literal[404]


class Unauthorized(ClientError):
    code: Literal[401]


class Forbidden(ClientError):
    code: Literal[403]



import uuid
from typing import cast

from ape_pie import APIClient

from openklant_client._resources.base import ResourceMixin
from openklant_client.types.pagination import PaginatedResponseBody
from openklant_client.types.resources.partij import (
    CreatePartijContactpersoonData,
    CreatePartijOrganisatieData,
    CreatePartijPersoonData,
    PartialUpdatePartijData,
    Partij,
    PartijListParams,
    PartijRetrieveParams,
)


class PartijResource(ResourceMixin):
    http_client: APIClient
    base_path: str = "partijen"

    def list(
        self, *, params: PartijListParams | None = None
    ) -> PaginatedResponseBody[Partij]:
        response = self._get(self.base_path, params=params)
        return cast(PaginatedResponseBody[Partij], self.process_response(response))

    def retrieve(
        self, /, uuid: str | uuid.UUID, *, params: PartijRetrieveParams | None = None
    ) -> Partij:
        response = self._get(f"{self.base_path}/{str(uuid)}", params=params)
        return cast(Partij, self.process_response(response))

    # Partij is polymorphic on "soortPartij", with varying fields for
    # Persoon, Organisatie and ContactPersoon
    def create_organisatie(self, *, data: CreatePartijOrganisatieData) -> Partij:
        return self._create(data=data)

    def create_persoon(
        self,
        *,
        data: CreatePartijPersoonData,
    ) -> Partij:
        return cast(Partij, self._create(data=data))

    def create_contactpersoon(
        self,
        *,
        data: CreatePartijContactpersoonData,
    ) -> Partij:
        return cast(Partij, self._create(data=data))

    def _create(
        self,
        *,
        data: (
            CreatePartijPersoonData
            | CreatePartijOrganisatieData
            | CreatePartijContactpersoonData
        ),
    ) -> Partij:
        response = self._post(self.base_path, data=data)
        return cast(Partij, self.process_response(response))

    def partial_update(
        self, /, uuid: str | uuid.UUID, *, data: PartialUpdatePartijData
    ) -> Partij:
        response = self._patch(f"{self.base_path}/{str(uuid)}", data=data)
        return cast(Partij, self.process_response(response))



from typing import Literal, NotRequired, Required

from typing_extensions import TypedDict

from openklant_client.types.common import BooleanQueryParam, ForeignKeyRef
from openklant_client.types.iso_639_2 import LanguageCode


class CreateKlantContactData(TypedDict):
    nummer: NotRequired[str]
    kanaal: str
    onderwerp: str
    inhoud: NotRequired[str]
    indicatieContactGelukt: NotRequired[bool]
    taal: LanguageCode
    vertrouwelijk: bool
    plaatsgevondenOp: NotRequired[str]


class ListKlantContactParams(TypedDict, total=False):
    expand: list[
        Literal[
            "gingOverOnderwerpobjecten",
            "hadBetrokkenen",
            "hadBetrokkenen.digitaleAdressen",
            "hadBetrokkenen.wasPartij",
            "leiddeTotInterneTaken",
            "omvatteBijlagen",
        ]
    ]
    hadBetrokkene__url: str
    hadBetrokkene__uuid: str
    indicatieContactGelukt: BooleanQueryParam
    inhoud: str
    kanaal: str
    nummer: str
    onderwerp: str
    onderwerpobject__onderwerpobjectidentificatorCodeObjecttype: str
    onderwerpobject__onderwerpobjectidentificatorCodeRegister: str
    onderwerpobject__onderwerpobjectidentificatorCodeSoortObjectId: str
    onderwerpobject__onderwerpobjectidentificatorObjectId: str
    onderwerpobject__url: str
    onderwerpobject__uuid: str
    page: int
    plaatsgevondenOp: str
    vertrouwelijk: BooleanQueryParam
    wasOnderwerpobject__onderwerpobjectidentificatorCodeObjecttype: str
    wasOnderwerpobject__onderwerpobjectidentificatorCodeRegister: str
    wasOnderwerpobject__onderwerpobjectidentificatorCodeSoortObjectId: str
    wasOnderwerpobject__onderwerpobjectidentificatorObjectId: str
    wasOnderwerpobject__url: str
    wasOnderwerpobject__uuid: str


class RetrieveKlantContactParams(TypedDict, total=False):
    expand: list[
        Literal[
            "gingOverOnderwerpobjecten",
            "hadBetrokkenen",
            "hadBetrokkenen.digitaleAdressen",
            "hadBetrokkenen.wasPartij",
            "leiddeTotInterneTaken",
            "omvatteBijlagen",
        ]
    ]


class KlantContact(TypedDict):
    uuid: str
    url: str
    # onderwerpobjecten: list[ForeignKeyRef]
    gingOverOnderwerpobjecten: list[ForeignKeyRef]
    hadBetrokkenActoren: list[ForeignKeyRef]  # HAs more
    omvatteBijlagen: list[ForeignKeyRef]
    hadBetrokkenen: list[ForeignKeyRef]
    leiddeTotInterneTaken: list[ForeignKeyRef]

    nummer: str
    kanaal: str
    onderwerp: str
    inhoud: str | None
    indicatieContactGelukt: Required[bool | None]

    taal: LanguageCode
    vertrouwelijk: bool
    plaatsgevondenOp: str



from typing import Literal, NotRequired

from typing_extensions import TypedDict

from openklant_client.types.common import ForeignKeyRef

#
# Input
#

CodeObjecttype = Literal[
    "natuurlijk_persoon",
    "niet_natuurlijk_persoon",
    "overig",
    "vestiging",
]

CodeSoortObjectId = Literal[
    "bsn",
    "kvk_nummer",
    "overig",
    "rsin",
    "vestigingsnummer",
]

CodeRegister = Literal[
    "brp",
    "hr",
    "overig",
]


class PartijIdentificatorObject(TypedDict):
    codeObjecttype: CodeObjecttype
    codeSoortObjectId: CodeSoortObjectId
    codeRegister: CodeRegister
    objectId: str


class CreateIdentificeerdePartij(TypedDict):
    uuid: str


# For use in the Partij creation endpoint
class CreateEmbeddedPartijIdentificatorData(TypedDict):
    partijIdentificator: PartijIdentificatorObject
    anderePartijIdentificator: NotRequired[str]
    subIdentificatorVan: NotRequired[ForeignKeyRef]


class CreatePartijIdentificatorData(TypedDict):
    identificeerdePartij: CreateIdentificeerdePartij
    partijIdentificator: PartijIdentificatorObject
    anderePartijIdentificator: NotRequired[str]
    subIdentificatorVan: NotRequired[ForeignKeyRef]


class ListPartijIdentificatorenParams(TypedDict, total=False):
    page: int
    partijIdentificatorCodeObjecttype: CodeObjecttype
    partijIdentificatorCodeRegister: CodeRegister
    partijIdentificatorCodeSoortObjectId: CodeSoortObjectId
    partijIdentificatorObjectId: str


#
# Output
#


class IdentificerendePartij(TypedDict):
    uuid: str
    url: str


class PartijIdentificator(TypedDict):
    uuid: str
    identificeerdePartij: IdentificerendePartij
    partijIdentificator: PartijIdentificatorObject
    anderePartijIdentificator: NotRequired[str]
    subIdentificatorVan: ForeignKeyRef | None
